import bge
import mathutils
import math
import aud
import mutagen
import random
import os

import trail

MAPPING = {
    'left':'AKEY',
    'right':'DKEY',
    'boost': 'SPACEKEY'
}

def get_key(action_name):
    key_str = MAPPING[action_name]
    key_code = bge.events.__dict__[key_str]
    return bge.logic.keyboard.events[key_code] != 0
    

def init(cont):
    if 'COUNTER' not in cont.owner:
        cont.owner['COUNTER'] = 0
        bge.logic.addScene('HUD')
    cont.owner['COUNTER'] += 1
    if cont.owner['COUNTER'] == 2:
        cont.owner['GAME'] = Game(cont.owner)
        cont.script = __name__ + '.run'
    
def run(cont):
    cont.owner['GAME'].update()




class Game(object):
    def __init__(self, player_obj):
        self.scene = player_obj.scene
        self.player = Player(player_obj)
        self.music = Music(bge.logic.expandPath('//Tracks'))
        self.hud = HUD([s for s in bge.logic.getSceneList() if s.name == 'HUD'][0])
        self.map = Map(self.scene)

        self.player.on_death.append(self.end)

        self.counter = 0
        
    def update(self):
        self.player.update()
        self.map.update(self.player.position)
        self.hud.set_speed(self.player.current_speed * 2)
        self.hud.set_percent(self.music.percent)

        if self.music.percent >= 0.99999:
            self.music.end()
            for scene in bge.logic.getSceneList():
                scene.end()
            bge.logic.addScene('COMPLETE')

        self.counter += 0.3
        name_str = self.music.name[max(0, int(self.counter) - 80):int(self.counter)]
        self.hud.set_name(name_str)

    def end(self, *args):
        self.music.end()
        for scene in bge.logic.getSceneList():
            scene.end()
        bge.logic.addScene('FAILED')
        

class HUD(object):
    def __init__(self, scene):
        self.scene = scene

    def set_speed(self, val):
        self.scene.objects['SPEED'].text = "{:03}".format(int(val))

    def set_percent(self, val):
        self.scene.objects['PERCENT'].color = [val] * 4

    def set_name(self, name):
        self.scene.objects['NAME'].text = name

class Player(object):
    def __init__(self, player_obj):
        self.obj = player_obj
        self.speed = 100
        self.trail = trail.Trail(self.obj.children['TRAIL'])
        self.trail.length = 2
        self.on_death = list()

        self.obj.collisionCallbacks.append(self._hit)
        
    def update(self):
        self.speed += 0.02
        self.trail.update()
        
        thrust = mathutils.Vector([0, 1, 0])
        thrust.x += get_key('right')
        thrust.x -= get_key('left')
        thrust.length = self.speed
        thrust.length *= (1 + get_key('boost') * 2.0)
        
        self.obj.applyForce(thrust)
        
        vel = self.obj.worldLinearVelocity
        ori = [0, 0, 0]
        ang = math.atan2(vel.x, vel.y)
        ori[1] = ang
        ori[2] = -ang
        self.obj.worldOrientation = ori

    def _hit(self, hit_obj):
        for funct in self.on_death:
            funct(self)
        
    @property
    def position(self):
        return self.obj.worldPosition

    @property
    def current_speed(self):
        return self.obj.worldLinearVelocity.length



class Music(object):
    def __init__(self, path):
        self.device = aud.device()
        file_name = random.choice(os.listdir(path))
        self.file_path = os.path.join(path, file_name)
        self.music = aud.Factory(self.file_path).buffer()
        self.handle = self.device.play(self.music)
        self.handle.keep = True

        try:
            self.file_info = mutagen.File(self.file_path)
            self.name = str(self.file_info['TIT2']) + ' - ' +str(self.file_info['TPE1'])
        except:
            self.name = file_name

        at_end = False
        duration = 0.0
        prev_duration = 0.0
        while not at_end:
            duration += 1.0
            self.handle.position = duration
            if self.handle.position != duration:
                at_end = True

        self.handle.keep = False
        self.handle.position = 0.0
        self.duration = duration

    @property
    def percent(self):
        '''Returns how far through the song it is'''
        return self.handle.position / self.duration

    def end(self):
        self.handle.stop()

    
class Map(object):
    def __init__(self, scene):
        self.scene = scene
        self._pieces = list()
        
    def _add_piece(self, position):
        '''Creates a new obstacle'''
        obj = random.choice([o for o in self.scene.objectsInactive if 'PIECE' in o])
        new_obj = self.scene.addObject(obj)
        
        new_obj.worldPosition.x = int(position.x)
        new_obj.worldPosition.y = int(position.y)
        new_obj.worldPosition.z = -0.7
        
        offset = abs(position.x) / 30
        new_obj.localScale = [max(1, int(offset))]*2 + [0]
        
        
        self._pieces.append(new_obj)
        
    def update(self, position):
        if position.y < 100:
            return
        for piece in self._pieces[:]:
            if piece.worldPosition.y < position.y - 20.0:  # Remove pieces behind player
                self._pieces.remove(piece)
                piece.endObject()
            else:
                piece.localScale[2] = piece.localScale[2] * 0.9 + piece.localScale[1] * 0.1
                
        if len(self._pieces) < 50:
            pos = position.copy()
            pos.y += random.uniform(100, 500)
            pos.x += random.uniform(-200, 200)
            self._add_piece(pos)